import html
import logging
import re
from copy import deepcopy
from datetime import datetime
from textwrap import TextWrapper
from urllib.parse import quote_plus as urlquote_plus
from urllib.parse import urljoin

import dateparser
import lxml
from dateutil.parser import parse as dateutil_parse
from dateutil.tz import gettz
from lxml.cssselect import CSSSelector
from lxml.html.clean import Cleaner
from scrapy.loader import ItemLoader
from scrapy.loader.processors import Compose, Identity, Join, MapCompose, TakeFirst
from w3lib.html import remove_tags

from feeds.items import FeedEntryItem, FeedItem
from feeds.settings import get_feeds_settings

logger = logging.getLogger(__name__)

_lxml_cleaner = Cleaner(style=True)

# List of so-called empty elements in HTML.
# Source: https://developer.mozilla.org/en-US/docs/Glossary/Empty_element
EMPTY_ELEMENTS = [
    "area",
    "base",
    "br",
    "col",
    "embed",
    "hr",
    "img",
    "input",
    "keygen",
    "link",
    "meta",
    "param",
    "source",
    "track",
    "wbr",
]


def parse_datetime(date_time, loader_context):
    if isinstance(date_time, datetime):
        return date_time
    elif isinstance(date_time, str):
        try:
            return dateutil_parse(
                date_time.strip(),
                dayfirst=loader_context.get("dayfirst", False),
                yearfirst=loader_context.get("yearfirst", True),
                ignoretz=loader_context.get("ignoretz", False),
            )
        except ValueError:
            # If dateutil can't parse it, it might be a human-readable date.
            return dateparser.parse(date_time)
    else:
        raise ValueError("date_time must be datetime or a str.")


def apply_timezone(date_time, loader_context):
    if not date_time.tzinfo:
        # If date_time object is not aware, apply timezone from loader_context.
        # In case a timezone is not supplied, just assume UTC.
        date_time = date_time.replace(
            tzinfo=gettz(loader_context.get("timezone", "UTC"))
        )
    return date_time


def replace_regex(text, loader_context):
    for pattern, repl in loader_context.get("replace_regex", {}).items():
        text = re.sub(pattern, repl, text)

    return text


def build_tree(text, loader_context=None):
    base_url = loader_context.get("base_url", None) if loader_context else None
    tree = lxml.html.fragment_fromstring(text, create_parent="div", base_url=base_url)

    if base_url:
        # Workaround for https://bugs.launchpad.net/lxml/+bug/1576598.
        # FIXME: Remove this when a workaround is released.
        tree.getroottree().docinfo.URL = base_url

    # Scrapy expects an iterator which it unpacks and feeds to the next
    # function in the pipeline. trees are iterators but we don't want to them
    # to get unpacked so we wrap the tree in another iterator.
    return [tree]


def serialize_tree(tree, in_make_links=False):
    return lxml.html.tostring(tree, encoding="unicode")


def make_links_absolute(tree):
    if tree.base_url:
        # Make references in tags like <a> and <img> absolute.
        tree.make_links_absolute(handle_failures="ignore")
    return [tree]


def pullup_elems(tree, loader_context):
    for elem_child, parent_dist in loader_context.get("pullup_elems", {}).items():
        selector = CSSSelector(elem_child)
        for elem in selector(tree):
            parent = elem
            for _ in range(parent_dist):
                parent = parent.getparent()
            if parent is not None and parent.getparent() is not None:
                elem.tail = parent.tail
                parent.getparent().replace(parent, elem)
            else:
                logger.error(
                    'Could not find parent with distance {} for selector "{}".'.format(
                        parent_dist, elem_child
                    )
                )

    return [tree]


def replace_elems(tree, loader_context):
    for elem_sel, elem_repl in loader_context.get("replace_elems", {}).items():
        selector = CSSSelector(elem_sel)
        for elem in selector(tree):
            # If elem_repl is callable, call it to create a new element (or just modify
            # the old one).
            if callable(elem_repl):
                elem_new = elem_repl(elem)
            else:
                elem_new = elem_repl

            # The new element is None, just remove the old one.
            if elem_new is None:
                elem.drop_tree()
            else:
                if isinstance(elem_new, str):
                    # The new element is a string, create a proper element out of it.
                    elem_new = lxml.html.fragment_fromstring(elem_new)
                else:
                    # Create a copy of elem_new in case the element should be used as a
                    # replacement more than once.
                    elem_new = deepcopy(elem_new)
                # Take care to preserve the tail of the old element.
                elem_new.tail = elem.tail
                elem.getparent().replace(elem, elem_new)

    return [tree]


def remove_elems(tree, loader_context):
    remove_elems = []

    settings = get_feeds_settings()
    remove_images = settings.getbool("FEEDS_CONFIG_REMOVE_IMAGES")
    if remove_images:
        remove_elems += ["img"]

    # Remove tags.
    for elem_sel in loader_context.get("remove_elems", []) + remove_elems:
        selector = CSSSelector(elem_sel)
        for elem in selector(tree):
            elem.drop_tree()

    for elem_sel in loader_context.get("remove_elems_xpath", []):
        for elem in tree.xpath(elem_sel):
            elem.drop_tree()

    return [tree]


def change_attribs(tree, loader_context):
    # Change attrib names.
    for elem_sel, attribs in loader_context.get("change_attribs", {}).items():
        selector = CSSSelector(elem_sel)
        for elem in selector(tree):
            for attrib in elem.attrib.keys():
                if attrib in attribs:
                    old_attrib_value = elem.attrib.pop(attrib)
                    if attribs[attrib] is not None:
                        # If attribs[attrib] is None, attrib is removed instead of
                        # renamed.
                        elem.attrib[attribs[attrib]] = old_attrib_value
    return [tree]


def change_tags(tree, loader_context):
    # Change tag names.
    for elem_sel, elem_tag in loader_context.get("change_tags", {}).items():
        selector = CSSSelector(elem_sel)
        for elem in selector(tree):
            elem.tag = elem_tag

    return [tree]


def cleanup_html(tree, loader_context):
    # tree.iter() iterates over the tree including the root node.
    for elem in tree.iter():
        # Remove class and id attribute from all elements which are not needed
        # in the feed.
        elem.attrib.pop("class", None)
        elem.attrib.pop("id", None)
        # Delete data- attributes that have no general meaning.
        for attrib in list(elem.attrib.keys()):
            if attrib.startswith("data-"):
                elem.attrib.pop(attrib)

    return [tree]


def lxml_cleaner(tree):
    _lxml_cleaner(tree)
    return [tree]


def convert_footnotes(tree, loader_context):
    # Convert footnotes.
    for elem_sel in loader_context.get("convert_footnotes", []):
        selector = CSSSelector(elem_sel)
        for elem in selector(tree):
            if elem.text is not None:
                elem.tag = "small"
                elem.text = " ({})".format(elem.text.strip())

    return [tree]


def convert_iframes(tree, loader_context):
    """Convert iframes to divs with links to its src.

    convert_iframes() is called after remove_elems() so that unwanted iframes can be
    eliminated first.
    """
    base_url = loader_context.get("base_url", None) if loader_context else None
    selector = CSSSelector("iframe")
    for elem in selector(tree):
        if "src" not in elem.attrib:
            continue
        url = urljoin(base_url, elem.attrib.pop("src"))
        elem_new = lxml.html.fragment_fromstring(
            '<div><a href="{url}">{url}</a></div>'.format(url=url)
        )
        elem_new.tail = elem.tail
        elem.getparent().replace(elem, elem_new)

    return [tree]


def flatten_tree(tree):
    # Post-order traversal.
    for child in tree.iterchildren():
        flatten_tree(child)

    # Points to the first child if tree has a child and it's the only child or None
    # otherwise.
    only_child = list(tree)[0] if len(tree) == 1 else None
    if (
        tree.tag not in EMPTY_ELEMENTS
        and (tree.text is None or tree.text.strip() == "")
        and len(tree) == 0
        and tree.getparent() is not None
    ):
        # Remove elements which don't have a text and are not supposed to be empty.
        tree.drop_tree()
        return None
    elif (
        only_child is not None
        and (tree.text is None or tree.text.strip() == "")
        and only_child.tag == tree.tag
        and tree.getparent() is not None
    ):
        # Replace tree with child if there is only one child and it has the same tag.

        # Preserve both tails.
        new_tail = (only_child.tail or "") + (tree.tail or "")
        if not new_tail:
            new_tail = None
        only_child.tail = new_tail

        tree.getparent().replace(tree, only_child)

    return [tree]


def skip_empty_tree(tree):
    if tree.text:
        # Has a text.
        return [tree]

    if len(tree):
        # Has children.
        return [tree]

    return None


def skip_none(value):
    """Skip values that are None immediately."""
    if value is not None:
        return value

    return None


def skip_false(value):
    """
    Skip values that evaluate to False.

    Scrapy only skips values that are None by default. In feeds we want to
    tighten that policy and also skip empty strings, False and everything else
    that evaluates to False.
    """
    if value:
        return value

    return None


def truncate_tree(tree, limit, drop=False):
    """Truncates the tree in-place so that its text will not exceed limit characters.

    If the limit is exceeded, words are dropped and "..." is added to the end. The
    difference between the actual length of the text and the limit is returned.
    If drop is True, the tree itself will be removed from its parent.
    """

    def _truncate(text, width):
        shorten = len(text) > width
        if shorten:
            # shorten() removes trailing spaces, so we use TextWrapper directly.
            text = TextWrapper(
                width=width, max_lines=1, placeholder="...", drop_whitespace=False
            ).fill(text)
        return (text, width - len(text), shorten)

    remaining = limit

    if remaining < 10 or drop:
        # This will effectively remove all children and the tail as well.
        tree.getparent().remove(tree)
        truncated = True
    else:
        truncated = False

        # Try to truncate the text of the tree.
        if tree.text:
            tree.text, remaining, truncated = _truncate(tree.text, remaining)

        # If there was not enough text but there are children, try to truncate the
        # children.
        for child in tree:
            # drop children if truncation was already done.
            remaining, truncated = truncate_tree(child, remaining, drop=truncated)

        # If there is a tail either truncate it or remove it.
        if tree.tail:
            if remaining >= 10:
                tree.tail, remaining, truncated = _truncate(tree.tail, remaining)
            else:
                tree.tail = None

    return remaining, truncated


def truncate_text(text):
    settings = get_feeds_settings()
    truncate_words = settings.getint("FEEDS_CONFIG_TRUNCATE_WORDS")
    if truncate_words and truncate_words > 0:
        tree = build_tree(text)[0]
        # Assume that a word has 5 characters on average.
        truncate_tree(tree, truncate_words * 5)
        text = serialize_tree(tree)
    return text


class BaseItemLoader(ItemLoader):
    # Defaults
    # Unescape twice to get rid of &amp;&xxx; encoding errors.
    default_input_processor = MapCompose(
        skip_none, str.strip, skip_false, html.unescape, html.unescape
    )
    default_output_processor = TakeFirst()

    # Join first two elements on ": " and the rest on " - ".
    title_out = Compose(lambda t: [": ".join(t[:2])] + t[2:], Join(" - "))

    updated_in = MapCompose(skip_false, parse_datetime, apply_timezone)

    author_name_out = Join(", ")

    # Optional
    path_in = MapCompose(urlquote_plus)
    path_out = Identity()


class FeedItemLoader(BaseItemLoader):
    default_item_class = FeedItem


class FeedEntryItemLoader(BaseItemLoader):
    default_item_class = FeedEntryItem

    # Field specific
    content_text_in = MapCompose(skip_false, str.strip, remove_tags)
    content_text_out = Join("\n")

    content_html_in = MapCompose(
        skip_false,
        replace_regex,
        build_tree,
        convert_footnotes,
        pullup_elems,
        replace_elems,
        remove_elems,
        change_attribs,
        change_tags,
        cleanup_html,
        convert_iframes,
        lxml_cleaner,
        flatten_tree,
        skip_empty_tree,
        make_links_absolute,
        serialize_tree,
    )
    content_html_out = Compose(Join(), truncate_text)

    # Use sorted to keep the output stable.
    category_out = Compose(set, sorted)

    enclosure_in = Identity()
    enclosure_out = Identity()



import configparser
import logging
import os

from scrapy.utils.project import get_project_settings

logger = logging.getLogger(__name__)

_SETTINGS = None


def load_feeds_settings(file_):
    settings = get_project_settings()
    set_feeds_settings(settings)
    if not file_:
        config_file_path = settings.get("FEEDS_CONFIG_FILE")
        try:
            file_ = open(config_file_path, "r")
        except IOError:
            logger.info("Could not load config file from {}!".format(config_file_path))
            return settings

    logger.debug("Parsing configuration file {} ...".format(file_.name))
    # Parse configuration file and store result under FEEDS_CONFIG of scrapy's
    # settings API.
    config = configparser.ConfigParser()
    config.read_file(file_)
    feeds_config = {s: dict(config.items(s)) for s in config.sections()}

    for key, value in feeds_config["feeds"].items():
        settings.set("FEEDS_CONFIG_{}".format(key.upper()), value)

    del feeds_config["feeds"]

    for spider in feeds_config.keys():
        spider_key = spider.replace(".", "_").upper()
        for key, value in feeds_config[spider].items():
            settings.set("FEEDS_SPIDER_{}_{}".format(spider_key, key.upper()), value)

    # Mapping of feeds config section to setting names.
    feeds_cfgfile_mapping = {
        "USER_AGENT": (config.get, "useragent", str),
        "LOG_LEVEL": (config.get, "loglevel", str),
        "HTTPCACHE_ENABLED": (config.getboolean, "cache_enabled", bool),
        "HTTPCACHE_DIR": (config.get, "cache_dir", os.path.expanduser),
        "HTTPCACHE_EXPIRATION_SECS": (
            config.getint,
            "cache_expires",
            lambda e: e * 24 * 60 * 60,
        ),
    }
    for key, value in feeds_cfgfile_mapping.items():
        setting_value = value[0]("feeds", value[1], fallback=None)
        if setting_value is not None:
            settings.set(key, value[2](setting_value))

    file_.close()

    return settings


def get_feeds_settings():
    return _SETTINGS


def set_feeds_settings(settings):
    global _SETTINGS
    _SETTINGS = settings



from setuptools import find_packages, setup

with open("README.rst") as f:
    long_description = f.read()

setup(
    name="PyFeeds",
    version="2020.5.16",
    description="DIY Atom feeds in times of social media and paywalls",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    author="Florian Preinstorfer, Lukas Anzinger",
    author_email="florian@nblock.org, lukas@lukasanzinger.at",
    url="https://github.com/PyFeeds/PyFeeds",
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        "Click>=6.6",
        "Scrapy>=1.6",
        "bleach>=1.4.3",
        "dateparser>=0.5.1",
        "feedparser",
        "lxml>=3.5.0",
        "python-dateutil>=2.7.3",
        "pyxdg>=0.26",
        "readability-lxml>=0.7",
        "scrapy-inline-requests",
    ],
    extras_require={
        "docs": ["sphinx", "sphinx_rtd_theme"],
        "style": [
            "black",
            "doc8",
            "flake8",
            "isort",
            "pygments",
            "restructuredtext_lint",
        ],
        "test": ["pytest"],
    },
    entry_points={"console_scripts": ["feeds=feeds.cli:main"]},
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Framework :: Scrapy",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Internet :: WWW/HTTP",
    ],
)



import re

import scrapy

from feeds.loaders import FeedEntryItemLoader
from feeds.spiders import FeedsXMLFeedSpider
from feeds.utils import generate_feed_header


class ArsTechnicaComSpider(FeedsXMLFeedSpider):
    name = "arstechnica.com"
    itertag = "item"

    def start_requests(self):
        channels = self.settings.get("FEEDS_SPIDER_ARSTECHNICA_COM_CHANNELS")
        if channels:
            channels = set(channels.split())
        else:
            channels = {"features"}

        for channel in channels:
            yield scrapy.Request(
                "http://feeds.{}/arstechnica/{}".format(self.name, channel),
                meta={"path": channel, "dont_cache": True},
            )

        self._channels = channels

    def feed_headers(self):
        for channel in self._channels:
            yield generate_feed_header(
                title="Ars Technica: {}".format(channel.title()),
                link="https://{}".format(self.name),
                path=channel,
                icon=(
                    "https://cdn.arstechnica.net/wp-content/uploads/2016/10/"
                    + "cropped-ars-logo-512_480-32x32.png"
                ),
                logo=(
                    "https://cdn.arstechnica.net/wp-content/themes/ars-mobile/assets/"
                    + "images/material-ars.png"
                ),
            )

    def parse_node(self, response, node):
        link = node.xpath("link/text()").extract_first()
        il = FeedEntryItemLoader()
        il.add_value("title", node.xpath("title/text()").extract_first())
        il.add_value("updated", node.xpath("pubDate/text()").extract_first())
        il.add_value("category", node.xpath("category/text()").extract())
        return scrapy.Request(
            link,
            self._parse_article,
            cookies={"view": "mobile"},
            meta={"il": il, "path": response.meta["path"], "first_page": True},
        )

    @staticmethod
    def _div_to_img(elem):
        elem.tag = "img"
        url = re.search(r"url\('([^']+)'\)", elem.attrib["style"]).group(1)
        elem.attrib["src"] = url
        elem.attrib["style"] = None
        return elem

    def _parse_article(self, response):
        remove_elems = [
            ".caption-credit",
            ".gallery-image-credit",
            "#social-left",
            "ul.toc",
            "h3:contains('Table of Contents')",
            "br",
            ".sidebar:contains('Further Reading')",
            ".credit",
        ]
        change_tags = {".sidebar": "blockquote", "aside": "blockquote"}
        replace_elems = {"div.image": self._div_to_img}
        il = FeedEntryItemLoader(
            response=response,
            parent=response.meta["il"],
            remove_elems=remove_elems,
            replace_elems=replace_elems,
            change_tags=change_tags,
        )
        if response.meta.get("first_page", False):
            il.add_value("link", response.url)
            il.add_css("author_name", ".byline a span ::text")
            il.add_css("content_html", "header h2")
            il.add_value("path", response.meta["path"])
        il.add_css("content_html", ".article-content")
        if response.css(".next"):
            return scrapy.Request(
                response.css(".numbers a::attr(href)").extract()[-1],
                self._parse_article,
                meta={"il": il, "path": response.meta["path"]},
            )
        else:
            return il.load_item()



import scrapy

from feeds.loaders import FeedEntryItemLoader
from feeds.spiders import FeedsSpider


class VerbraucherrechtAtSpider(FeedsSpider):
    name = "verbraucherrecht.at"
    start_urls = ["https://verbraucherrecht.at/cms/index.php?id=198"]

    feed_title = "Verbraucherrecht"
    feed_subtitle = "Neuigkeiten rund um Konsumentenschutz und Verbraucherrechte"
    feed_author_name = "Verein für Konsumenteninformation"
    feed_link = "https://{}".format(name)
    feed_logo = "https://{}/cms/fileadmin/imag/logo.gif".format(name)

    def parse(self, response):
        # Fetch only the current news page and deliberately ignore old news.
        for link in response.xpath(
            '//div[@class="news-list-container"]//h2/a/@href'
        ).extract():
            yield scrapy.Request(response.urljoin(link), self.parse_item)

    def parse_item(self, response):
        il = FeedEntryItemLoader(
            response=response,
            base_url="{}/cms/".format(self.feed_link),
            timezone="Europe/Vienna",
            remove_elems=[".news-latest-date", ".news-single-rightbox", "hr", "h7"],
            remove_elems_xpath=[
                '//div[@class="news-single-item"]/b[1]',
                '//div[@class="news-single-item"]/br[1]',
            ],
            dayfirst=True,
        )

        il.add_value(
            "title", response.xpath("//head/title/text()").re_first(r"::: (.*)")
        )

        il.add_value("link", response.url)

        il.add_value(
            "updated",
            response.xpath('//div[@class="news-single-rightbox"]').re_first(
                r"(\d{2}\.\d{2}\.\d{4})"
            ),
        )

        il.add_value(
            "author_name",
            response.xpath('//head/meta[@name="publisher"]/@content').re_first(
                "recht.at, (.*);"
            ),
        )
        il.add_xpath("author_name", '//head/meta[@name="author"]/@content')
        il.add_value("author_name", self.name)

        il.add_xpath("author_email", '//head/meta[@name="reply-to"]/@content')

        il.add_css("content_html", ".news-single-item h7 font strong")
        il.add_css("content_html", ".news-single-item")

        return il.load_item()



import json

import scrapy

from feeds.loaders import FeedEntryItemLoader
from feeds.spiders import FeedsSpider
from feeds.utils import generate_feed_header


class SpotifyComSpider(FeedsSpider):
    name = "spotify.com"

    def start_requests(self):
        self._shows = self.settings.get("FEEDS_SPIDER_SPOTIFY_COM_SHOWS")
        if self._shows:
            self._shows = set(self._shows.split())
        else:
            self.logger.error("No shows given!")
            return

        self._market = self.settings.get("FEEDS_SPIDER_SPOTIFY_COM_MARKET")
        if not self._market:
            self.logger.error("Please specify a market (country code) in the config!")
            return

        yield scrapy.Request(
            "https://open.spotify.com/get_access_token?"
            + "reason=transport&productType=web_player",
            meta={"dont_cache": True},  # Only valid for a very short time.
        )

    def parse(self, response):
        response = json.loads(response.text)
        access_token = response["accessToken"]
        for show in self._shows:
            yield scrapy.Request(
                f"https://api.spotify.com/v1/shows/{show}?market={self._market}",
                self._parse_show,
                meta={"spotify_show": show, "dont_cache": True},
                headers={"Authorization": f"Bearer {access_token}"},
            )

    def feed_headers(self):
        return []

    def _parse_show(self, response):
        result = json.loads(response.text)

        yield generate_feed_header(
            title=result["name"],
            link=result["external_urls"]["spotify"],
            icon=result["images"][-1]["url"],
            logo=result["images"][0]["url"],
            path=response.meta["spotify_show"],
        )

        for episode in result["episodes"]["items"]:
            il = FeedEntryItemLoader()
            il.add_value("link", episode["external_urls"]["spotify"])
            il.add_value("updated", episode["release_date"])
            il.add_value("title", episode["name"])
            il.add_value("content_html", episode["description"])
            il.add_value("path", response.meta["spotify_show"])
            yield il.load_item()



import os
import uuid
from datetime import datetime, timezone

from scrapy import signals
from scrapy.exceptions import DropItem

from feeds.exporters import AtomExporter
from feeds.items import FeedEntryItem


class AtomAutogenerateFieldsPipeline(object):
    """Autogenerate missing fields in case they are missing."""

    def process_item(self, item, spider):
        if "id" not in item:
            if "link" in item:
                item["id"] = uuid.uuid5(uuid.NAMESPACE_DNS, item["link"]).urn
            else:
                raise DropItem(
                    "A link is required to autogenerate the feed "
                    "id for: {}".format(item)
                )

        if "title" not in item:
            # Having a title is mandatory, so we use an empty string if none
            # is set.
            item["title"] = ""

        if isinstance(item, FeedEntryItem) and "updated" not in item:
            if "link" in item:
                item["updated"] = spider.cache.setdefault(
                    spider,
                    key="{}|updated".format(item["id"]),
                    default_obj=datetime.now(timezone.utc),
                )
            else:
                raise DropItem(
                    "A link is required to autogenerate the updated field "
                    "for: {}".format(item)
                )

        return item


class AtomCheckRequiredFieldsPipeline(object):
    """Check presence of required fields."""

    def process_item(self, item, spider):
        def raise_if_missing(name, item):
            if name not in item:
                raise DropItem(
                    'The required field "{}" is missing in: {}.'.format(name, item)
                )

        # Required fields for all items
        for required in ("id", "title", "link"):
            raise_if_missing(required, item)

        # Required fields for FeedEntryItems
        if isinstance(item, FeedEntryItem):
            for required in ("updated",):
                raise_if_missing(required, item)

        return item


class AtomExportPipeline(object):
    """Export items as atom feeds."""

    def __init__(self, output_path, output_url):
        self._output_path = os.path.expanduser(output_path)
        self._output_url = output_url
        self._exporters = {}

    @classmethod
    def from_crawler(cls, crawler):
        output_path = crawler.settings.get("FEEDS_CONFIG_OUTPUT_PATH")
        if not output_path:
            raise ValueError("output_path not set!")
        output_url = crawler.settings.get("FEEDS_CONFIG_OUTPUT_URL")
        pipeline = cls(output_path=output_path, output_url=output_url)
        crawler.signals.connect(pipeline.spider_opened, signals.spider_opened)
        crawler.signals.connect(pipeline.spider_closed, signals.spider_closed)
        return pipeline

    def spider_opened(self, spider):
        self._exporters[spider] = AtomExporter(
            self._output_path, self._output_url, spider.name
        )
        self._exporters[spider].start_exporting()

    def spider_closed(self, spider):
        # Add feed header(s) at the end so they can be dynamic.
        for feed_header in spider.feed_headers():
            self._exporters[spider].export_item(feed_header)
        self._exporters[spider].finish_exporting()
        self._exporters.pop(spider)

    def process_item(self, item, spider):
        self._exporters[spider].export_item(item)
        return item
